-- IMAGE_NT_HEADERS
-- (c) FSC, 2007

local IMAGE_NT_SIGNATURE = 0x00004550
local IMAGE_NUMBEROF_DIRECTORY_ENTRIES = 16

--
-- show Machine type in IMAGE_FILE_HEADER
--
local IMAGE_FILE_MACHINE =
{
    IMAGE_FILE_MACHINE_UNKNOWN      =   0,
    IMAGE_FILE_MACHINE_I386         =   0x014c,
    IMAGE_FILE_MACHINE_R3000        =   0x0162,
    IMAGE_FILE_MACHINE_R4000        =   0x0166,
    IMAGE_FILE_MACHINE_R10000       =   0x0168,
    IMAGE_FILE_MACHINE_WCEMIPSV2    =   0x0169,
    IMAGE_FILE_MACHINE_ALPHA        =   0x0184,
    IMAGE_FILE_MACHINE_POWERPC      =   0x01F0,
    IMAGE_FILE_MACHINE_SH3          =   0x01a2,
    IMAGE_FILE_MACHINE_SH3E         =   0x01a4,
    IMAGE_FILE_MACHINE_SH4          =   0x01a6,
    IMAGE_FILE_MACHINE_ARM          =   0x01c0,
    IMAGE_FILE_MACHINE_THUMB        =   0x01c2,
    IMAGE_FILE_MACHINE_IA64         =   0x0200,
    IMAGE_FILE_MACHINE_MIPS16       =   0x0266,
    IMAGE_FILE_MACHINE_MIPSFPU      =   0x0366,
    IMAGE_FILE_MACHINE_MIPSFPU16    =   0x0466,
    IMAGE_FILE_MACHINE_ALPHA64      =   0x0284
}

local function Machine()
    StartType("WORD", "Machine")
    OneOf(GetUI16(), IMAGE_FILE_MACHINE, "0x%04x")
    EndType()
end

--
-- show Characteristics
--
local CHARACTERISTICS =
{
    IMAGE_FILE_RELOCS_STRIPPED          =   0x0001,
    IMAGE_FILE_EXECUTABLE_IMAGE         =   0x0002,
    IMAGE_FILE_LINE_NUMS_STRIPPED       =   0x0004,
    IMAGE_FILE_LOCAL_SYMS_STRIPPED      =   0x0008,
    IMAGE_FILE_AGGRESIVE_WS_TRIM        =   0x0010,
    IMAGE_FILE_LARGE_ADDRESS_AWARE      =   0x0020,
    IMAGE_FILE_BYTES_REVERSED_LO        =   0x0080,
    IMAGE_FILE_32BIT_MACHINE            =   0x0100,
    IMAGE_FILE_DEBUG_STRIPPED           =   0x0200,
    IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP  =   0x0400,
    IMAGE_FILE_NET_RUN_FROM_SWAP        =   0x0800,
    IMAGE_FILE_SYSTEM                   =   0x1000,
    IMAGE_FILE_DLL                      =   0x2000,
    IMAGE_FILE_UP_SYSTEM_ONLY           =   0x4000,
    IMAGE_FILE_BYTES_REVERSED_HI        =   0x8000
}

local function Characteristics()
    StartType("WORD", "Characteristics")
    local val = GetUI16()
    PrintNumber(BitSetFmt(val, CHARACTERISTICS, "0x%04x"))
    PrintMLComment(string.format("0x%04x", val))
    EndType()
end

--
-- view raw and decoded TimeDateStamp
--
local function TimeDateStamp()
    local x = PeekUnixTime()
    local TDComment = "?????"
    if x then
        local MonthTable = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" }
        TDComment = string.format("%d %s %d %d:%02d:%02d", x.Day, MonthTable[x.Month], x.Year, x.Hour, x.Minute, x.Second)
    end
    Hex()
    DWORD("TimeDateStamp", TDComment)
end

--
-- IMAGE_FILE_HEADER
--
local function _IMAGE_FILE_HEADER()
    TypeWidth = 6
    NameWidth = 21
  
    Machine()
    Dec()  
    WORD "NumberOfSections"
    Hex()
    TimeDateStamp()
    local Pts = PeekUI32()
    if Pts == 0 then
        Dec()
        DWORD("PointerToSymbolTable", "no symbol table")
    else
        Hex()
        DWORD "PointerToSymbolTable"
    end
    DecHex()
    DWORD "NumberOfSymbols"
    local OHSize = PeekUI16()
    WORD "SizeOfOptionalHeader"
    Characteristics()
    return OHSize
end

IMAGE_FILE_HEADER = struct("IMAGE_FILE_HEADER", _IMAGE_FILE_HEADER)

--
-- IMAGE_OPTIONAL_HEADER.Subsystem
--
local SUBSYSTEM = {
    IMAGE_SUBSYSTEM_UNKNOWN         =   0,
    IMAGE_SUBSYSTEM_NATIVE          =   1,
    IMAGE_SUBSYSTEM_WINDOWS_GUI     =   2,
    IMAGE_SUBSYSTEM_WINDOWS_CUI     =   3,
    IMAGE_SUBSYSTEM_OS2_CUI         =   5,
    IMAGE_SUBSYSTEM_POSIX_CUI       =   7,
    IMAGE_SUBSYSTEM_NATIVE_WINDOWS  =   8,
    IMAGE_SUBSYSTEM_WINDOWS_CE_GUI  =   9
}

local function Subsystem()
    StartType("WORD", "Subsystem")
    OneOf(GetUI16(), SUBSYSTEM, "0x%04x")
    EndType()
end

--
-- IMAGE_OPTIONAL_HEADER.DllCharacteristics
--
local function DllCharacteristics()
    StartType("WORD", "DllCharacteristics")
    local val = GetUI16()
    PrintNumber(BitSetFmt(val, { IMAGE_DLLCHARACTERISTICS_WDM_DRIVER = 0x2000 }, "0x%04x"))
    if val ~= 0 then PrintMLComment(string.format("0x%04x", val)) end
    EndType()
end

--
-- IMAGE_DATA_DIRECTORY
--
local function _IMAGE_DATA_DIRECTORY()
    TypeWidth = 6
    NameWidth = 15
    Hex()
    DWORD "VirtualAddress"
    HexDec()
    DWORD "Size"
end

IMAGE_DATA_DIRECTORY = struct("IMAGE_DATA_DIRECTORY", _IMAGE_DATA_DIRECTORY)

--
-- IMAGE_OPTIONAL_HEADER
--
local function ShowOHMagic()
    StartType("WORD", "Magic")
    OneOf(GetUI16(), { IMAGE_NT_OPTIONAL_HDR32_MAGIC = 0x10b, IMAGE_NT_OPTIONAL_HDR64_MAGIC = 0x20b, IMAGE_ROM_OPTIONAL_HDR_MAGIC = 0x107 }, "0x%04x")
    EndType()
end

local function _IMAGE_OPTIONAL_HEADER()
    TypeWidth = 6
    NameWidth = 28
    
    BlockComment("\rStandard fields\r\r")
    ShowOHMagic()
    Dec()
    BYTE "MajorLinkerVersion"
    BYTE "MinorLinkerVersion"
    HexDec()
    DWORD "SizeOfCode"
    DWORD "SizeOfInitializedData"
    DWORD "SizeOfUninitializedData"
    Hex()
    DWORD "AddressOfEntryPoint"
    DWORD "BaseOfCode"
    DWORD "BaseOfData"
    
    CrLf()
    BlockComment("\rNT additional fields\r\r")
    DWORD "ImageBase"

    HexDec()
    DWORD "SectionAlignment"
    DWORD "FileAlignment"
    
    Dec()
    WORD "MajorOperatingSystemVersion"
    WORD "MinorOperatingSystemVersion"
    WORD "MajorImageVersion"
    WORD "MinorImageVersion"
    WORD "MajorSubsystemVersion"
    WORD "MinorSubsystemVersion"
    DWORD "Win32VersionValue"
    
    HexDec()
    DWORD "SizeOfImage"
    DWORD "SizeOfHeaders"
    Hex()
    DWORD "CheckSum"
    Subsystem()
    DllCharacteristics()
    
    HexDec()
    DWORD "SizeOfStackReserve"
    DWORD "SizeOfStackCommit"
    DWORD "SizeOfHeapReserve"
    DWORD "SizeOfHeapCommit"

    Hex()
    DWORD("LoaderFlags", "obsolete")
    Dec()
    DWORD "NumberOfRvaAndSizes"
    
    CrLf()
    NameWidth = 16
    array("IMAGE_DATA_DIRECTORY", "DataDirectory", IMAGE_DATA_DIRECTORY, IMAGE_NUMBEROF_DIRECTORY_ENTRIES)

--    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
end

IMAGE_OPTIONAL_HEADER = struct("IMAGE_OPTIONAL_HEADER", _IMAGE_OPTIONAL_HEADER)

--
-- IMAGE_NT_HEADERS
--
local _IMAGE_NT_HEADERS = function()
    if PeekUI32() ~= IMAGE_NT_SIGNATURE then
        Warning = "should be 'PE'"
        WarningBox("This is not a 'PE' file")
    end
    Char()
    DWORD("Signature", "IMAGE_NT_SIGNATURE")
    Hex()
    IMAGE_FILE_HEADER "FileHeader"
    IMAGE_OPTIONAL_HEADER "OptionalHeader"
end

IMAGE_NT_HEADERS = struct("IMAGE_NT_HEADERS", _IMAGE_NT_HEADERS)

